from .Blockchain import Chain, Account, Contract, BlockchainUtils
from .Cryptography import ModernCryptoUtils, ClassicalCryptoUtils
from .Misc import MiscUtils, MiscAdvancedUtils
from .PoW import PoWUtils



from web3 import Web3
from loguru import logger
from sys import exc_info
from json import dumps


class Chain():
    def __init__(self, RPCUrl: str, RequestParams: dict = None):
        from web3 import HTTPProvider
        from web3.middleware import geth_poa_middleware
        from time import time

        StartTime = time()
        self.Net = Web3(HTTPProvider(RPCUrl, request_kwargs=RequestParams))
        if self.Net.isConnected():
            FinishTime = time()
            Delay = round((FinishTime - StartTime) * 1000)
            logger.success(f"\n[ConnectToChain]Successfully connected to [{RPCUrl}]. [Delay] {Delay} ms")
            self.Net.middleware_onion.inject(geth_poa_middleware, layer=0)
            self.GetBasicInformation()
        else:
            logger.error(f"\n[ConnectToChain]Failed to connect to [{RPCUrl}].")
            raise Exception("Failed to connect to chain.")

    def GetBasicInformation(self) -> dict:
        try:
            ChainId = self.Net.eth.chainId
            BlockNumber = self.Net.eth.block_number
            GasPrice = Web3.fromWei(self.Net.eth.gas_price, "gwei")
            ClientVersion = self.Net.clientVersion
            logger.success(f"\n[BasicInformation]\n[ChainId]{ChainId}\n[BlockNumber]{BlockNumber}\n[GasPrice]{GasPrice} Gwei\n[ClientVersion]{ClientVersion}")
            return {"ChainId": ChainId, "BlockNumber": BlockNumber, "GasPrice": GasPrice, "ClientVersion": ClientVersion}
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[BasicInformation]Failed to get basic information.\n[ExceptionInformation]{ExceptionInformation}")
            return None

    def GetBlockInformation(self, BlockID="latest") -> dict:
        try:
            Info = self.Net.eth.get_block(BlockID)
            BlockHash = Info.hash.hex()
            BlockNumber = Info.number
            BlockTimeStamp = Info.timestamp
            BlockTransactionAmount = len(Info.transactions)
            logger.success(f"\n[BlockInformation][{BlockID}]\n[Hash]{BlockHash}\n[Number]{BlockNumber}\n[TimeStamp]{BlockTimeStamp}\n[TransactionAmount]{BlockTransactionAmount}")
            return {"BlockID": BlockID, "BlockHash": BlockHash, "BlockNumber": BlockNumber, "BlockTimeStamp": BlockTimeStamp, "BlockTransactionAmount": BlockTransactionAmount}
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[BlockInformation]Failed to get block [{BlockID}] information.\n[ExceptionInformation]{ExceptionInformation}")
            return None

    def GetTransactionByHash(self, TransactionHash: str) -> dict:
        try:
            Info = self.Net.eth.get_transaction(TransactionHash)
            BlockNumber = Info.blockNumber
            TransactionIndex = Info.transactionIndex
            From = Info["from"]
            To = Info.to
            GasPrice = Info.gasPrice
            Gas = Info.gas
            Nonce = Info.nonce
            Value = Info.value
            InputData = Info.input
            if GasPrice != None:
                TransactionType = "Traditional"
                GasPrice = Web3.fromWei(GasPrice, "gwei")
                logger.info(
                    f"\n[TransactionInformation][{TransactionHash}]\n[TransactionType]{TransactionType}\n[BlockNumber]{BlockNumber} [TransactionIndex]{TransactionIndex}\n[From]{From}\n[To]{To}\n[GasPrice]{GasPrice} Gwei [Gas]{Gas}\n[Nonce]{Nonce} [Value]{Value}\n[InputData]{InputData}")
                return {"TransactionHash": TransactionHash, "TransactionType": TransactionType, "BlockNumber": BlockNumber, "TransactionIndex": TransactionIndex, "From": From, "To": To, "GasPrice": GasPrice, "Gas": Gas, "Nonce": Nonce, "Value": Value, "InputData": InputData}
            else:
                TransactionType = "EIP-1559"
                MaxFeePerGas = Web3.fromWei(Info.maxFeePerGas, "gwei")
                MaxPriorityFeePerGas = Web3.fromWei(Info.maxPriorityFeePerGas, "gwei")
                logger.info(
                    f"\n[TransactionInformation][{TransactionHash}]\n[TransactionType]{TransactionType}\n[BlockNumber]{BlockNumber} [TransactionIndex]{TransactionIndex}\n[From]{From}\n[To]{To}\n[MaxFeePerGas]{MaxFeePerGas} Gwei\n[MaxPriorityFeePerGas]{MaxPriorityFeePerGas} Gwei\n[Nonce]{Nonce} [Value]{Value} [Gas]{Gas}\n[InputData]{InputData}")
                return {"TransactionHash": TransactionHash, "TransactionType": TransactionType, "BlockNumber": BlockNumber, "TransactionIndex": TransactionIndex, "From": From, "To": To, "MaxFeePerGas": MaxFeePerGas, "MaxPriorityFeePerGas": MaxPriorityFeePerGas, "Nonce": Nonce, "Value": Value, "Gas": Gas, "InputData": InputData}
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[TransactionInformation]Failed to get transaction [{TransactionHash}] information.\n[ExceptionInformation]{ExceptionInformation}")
            return None

    def GetTransactionByBlockIdAndIndex(self, BlockID, TransactionIndex: int) -> dict:
        try:
            Info = self.Net.eth.get_transaction_by_block(BlockID, TransactionIndex)
            BlockNumber = Info.blockNumber
            TransactionHash = Info.hash.hex()
            From = Info["from"]
            To = Info.to
            GasPrice = Info.gasPrice
            Gas = Info.gas
            Nonce = Info.nonce
            Value = Info.value
            InputData = Info.input
            if GasPrice != None:
                TransactionType = "Traditional"
                GasPrice = Web3.fromWei(GasPrice, "gwei")
                logger.info(
                    f"\n[TransactionInformation][{BlockID}][{TransactionIndex}]\n[TransactionType]{TransactionType}\n[BlockNumber]{BlockNumber} [TransactionIndex]{TransactionIndex}\n[From]{From}\n[To]{To}\n[GasPrice]{GasPrice} Gwei [Gas]{Gas}\n[Nonce]{Nonce} [Value]{Value}\n[InputData]{InputData}")
                return {"TransactionHash": TransactionHash, "TransactionType": TransactionType, "BlockNumber": BlockNumber, "TransactionIndex": TransactionIndex, "From": From, "To": To, "GasPrice": GasPrice, "Gas": Gas, "Nonce": Nonce, "Value": Value, "InputData": InputData}
            else:
                TransactionType = "EIP-1559"
                MaxFeePerGas = Web3.fromWei(Info.maxFeePerGas, "gwei")
                MaxPriorityFeePerGas = Web3.fromWei(Info.maxPriorityFeePerGas, "gwei")
                logger.info(
                    f"\n[TransactionInformation][{BlockID}][{TransactionIndex}]\n[TransactionType]{TransactionType}\n[BlockNumber]{BlockNumber} [TransactionIndex]{TransactionIndex}\n[From]{From}\n[To]{To}\n[MaxFeePerGas]{MaxFeePerGas} Gwei\n[MaxPriorityFeePerGas]{MaxPriorityFeePerGas} Gwei\n[Nonce]{Nonce} [Value]{Value} [Gas]{Gas}\n[InputData]{InputData}")
                return {"TransactionHash": TransactionHash, "TransactionType": TransactionType, "BlockNumber": BlockNumber, "TransactionIndex": TransactionIndex, "From": From, "To": To, "MaxFeePerGas": MaxFeePerGas, "MaxPriorityFeePerGas": MaxPriorityFeePerGas, "Nonce": Nonce, "Value": Value, "Gas": Gas, "InputData": InputData}
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[TransactionInformation]Failed to get transaction [{BlockID}][{TransactionIndex}] information.\n[ExceptionInformation]{ExceptionInformation}")
            return None

    def GetBalance(self, Address: str) -> int:
        try:
            Balance = self.Net.eth.get_balance(Address)
            logger.success(f"\n[GetBalance][{Address}]\n[{Balance} Wei]<=>[{Web3.fromWei(Balance,'ether')} Ether]")
            return Balance
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[GetBalance]Failed to get [{Address}] balance.\n[ExceptionInformation]{ExceptionInformation}")
            return None

    def GetCode(self, Address: str) -> str:
        try:
            Code = self.Net.eth.get_code(Address).hex()
            logger.success(f"\n[GetCode][{Address}]\n[Code]{Code}")
            return Code
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[GetCode]Failed to get [{Address}] code.\n[ExceptionInformation]{ExceptionInformation}")
            return None

    def GetStorage(self, Address: str, Index: int) -> str:
        try:
            Data = self.Net.eth.get_storage_at(Address, Index).hex()
            logger.success(f"\n[GetStorage][{Address}][{Index}]\n[Hex][{Data}]<=>[Dec][{int(Data,16)}]")
            return Data
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[GetStorage]Failed to get [{Address}][{Index}] storage.\n[ExceptionInformation]{ExceptionInformation}")
            return None

    def DumpStorage(self, Address: str, Count: int) -> list:
        try:
            Data = [self.Net.eth.get_storage_at(Address, i).hex() for i in range(Count)]
            Temp = '\n'.join(Data)
            logger.info(f"\n[DumpStorage][{Address}][slot 0 ... {Count-1}]\n{Temp}")
            return Data
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[DumpStorage]Failed to dump [{Address}][0-{Count-1}] storages.\n[ExceptionInformation]{ExceptionInformation}")
            return None


class Account():
    def __init__(self, Chain: Chain, PrivateKey: str):
        try:
            self.Chain = Chain
            self.Net = Chain.Net
            AccountTemp = self.Net.eth.account.from_key(PrivateKey)
            self.Address = Web3.toChecksumAddress(AccountTemp.address)
            self.PrivateKey = AccountTemp.privateKey
            self.Net.eth.default_account = self.Address
            logger.success(f"\n[ImportAccount]Successfully import account [{self.Address}].")
            self.GetSelfBalance()
        except:
            ExceptionInformation = exc_info()
            logger.error(f"\n[ImportAccount]Failed to import account.\n[ExceptionInformation]{ExceptionInformation}")
            raise Exception("Failed to import account.")

    def GetSelfBalance(self) -> int:
        Balance = self.Chain.GetBalance(self.Address)
        if Balance == 0:
            logger.warning(f"\n[GetSelfBalance]Warning: This account's balance is insufficient to pay transactions fee.")
        return Balance

    def SendTransaction(self, To: str, Data: str, Value: int = 0, Gas: int = 10000000) -> dict:
        try:
            Txn = {
                "chainId": self.Net.eth.chainId,
                "from": self.Address,
                "to": Web3.toChecksumAddress(To),
                "nonce": self.Net.eth.get_transaction_count(self.Address),
                "value": Value,
                "gasPrice": round(self.Net.eth.gas_price * 1.2),
                "gas": Gas,
                "data": Data,
            }
            SignedTxn = self.Net.eth.account.sign_transaction(Txn, self.PrivateKey)
            TransactionHash = self.Net.eth.send_raw_transaction(SignedTxn.rawTransaction).hex()
            Txn["gasPrice"] = f'{Web3.fromWei(Txn["gasPrice"],"gwei")} Gwei'
            logger.info(f"\n[SendTransaction][Traditional]\n[TransactionHash]{TransactionHash}\n[Txn]{dumps(Txn, indent=2)}")
            TransactionReceipt = self.Net.eth.wait_for_transaction_receipt(TransactionHash, timeout=120)
            Status = TransactionReceipt.status
            if Status:
                BlockNumber = TransactionReceipt.blockNumber
                GasUsed = TransactionReceipt.gasUsed
                Logs = TransactionReceipt.logs
                logger.success(f"\n[ConfirmTransaction][Traditional][Success]\n[TransactionHash]{TransactionHash}\n[BlockNumber]{BlockNumber}\n[GasUsed]{GasUsed}\n[Logs]{Logs}")
                return {"Status": Status, "TransactionHash": TransactionHash, "BlockNumber": BlockNumber, "GasUsed": GasUsed, "Logs": Logs}
            else:
                logger.error(f"\n[ConfirmTransaction][Traditional][Fail]\n[TransactionHash]{TransactionHash}")
                return {"Status": Status, "TransactionHash": TransactionHash}
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[SendTransaction][Traditional]Failed to send transaction.\n[ExceptionInformation]{ExceptionInformation}")
            return None

    def SendTransactionByEIP1559(self, To: str, Data: str, Value: int = 0, Gas: int = 1000000) -> dict:
        try:
            Txn = {
                "chainId": self.Net.eth.chainId,
                "from": self.Address,
                "to": Web3.toChecksumAddress(To),
                "nonce": self.Net.eth.get_transaction_count(self.Address),
                "value": Value,
                "maxFeePerGas": self.Net.eth.max_priority_fee * 2,
                "maxPriorityFeePerGas": round(self.Net.eth.max_priority_fee * 1.2),
                "gas": Gas,
                "data": Data
            }
            SignedTxn = self.Net.eth.account.sign_transaction(Txn, self.PrivateKey)
            TransactionHash = self.Net.eth.send_raw_transaction(SignedTxn.rawTransaction).hex()
            Txn["gasPrice"] = f'{Web3.fromWei(Txn["gasPrice"],"gwei")} Gwei'
            logger.info(f"\n[SendTransaction][EIP-1559]\n[TransactionHash]{TransactionHash}\n[Txn]{dumps(Txn, indent=2)}")
            TransactionReceipt = self.Net.eth.wait_for_transaction_receipt(TransactionHash, timeout=120)
            Status = TransactionReceipt.status
            if Status:
                BlockNumber = TransactionReceipt.blockNumber
                GasUsed = TransactionReceipt.gasUsed
                Logs = TransactionReceipt.logs
                logger.success(f"\n[ConfirmTransaction][EIP-1559][Success]\n[TransactionHash]{TransactionHash}\n[BlockNumber]{BlockNumber}\n[GasUsed]{GasUsed}\n[Logs]{Logs}")
                return {"Status": Status, "TransactionHash": TransactionHash, "BlockNumber": BlockNumber, "GasUsed": GasUsed, "Logs": Logs}
            else:
                logger.error(f"\n[ConfirmTransaction][EIP-1559][Fail]\n[TransactionHash]{TransactionHash}")
                return {"Status": Status, "TransactionHash": TransactionHash}
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[SendTransaction][EIP-1559]Failed to send transaction.\n[ExceptionInformation]{ExceptionInformation}")
            return None

    def DeployContract(self, ABI: dict, Bytecode: str, Value: int = 0, *Arguments) -> dict:
        try:
            DeployingContract = self.Net.eth.contract(abi=ABI, bytecode=Bytecode)
            TransactionData = DeployingContract.constructor(*Arguments).buildTransaction({"value": Value})
            Txn = {
                "chainId": self.Net.eth.chainId,
                "from": self.Address,
                "nonce": self.Net.eth.get_transaction_count(self.Address),
                "value": TransactionData["value"],
                "gasPrice": round(self.Net.eth.gas_price * 1.2),
                "gas": TransactionData["gas"],
                "data": TransactionData["data"]
            }
            SignedTxn = self.Net.eth.account.sign_transaction(Txn, self.PrivateKey)
            TransactionHash = self.Net.eth.send_raw_transaction(SignedTxn.rawTransaction).hex()
            logger.info(f"\n[DeployContract][Traditional]\n[TransactionHash]{TransactionHash}\n[Txn]{dumps(Txn, indent=2)}")
            TransactionReceipt = self.Net.eth.wait_for_transaction_receipt(TransactionHash, timeout=120)
            Status = TransactionReceipt.status
            if Status:
                ContractAddress = TransactionReceipt.contractAddress
                BlockNumber = TransactionReceipt.blockNumber
                GasUsed = TransactionReceipt.gasUsed
                Logs = TransactionReceipt.logs
                logger.success(f"\n[ConfirmDeploy][Success]\n[ContractAddress]{ContractAddress}\n[TransactionHash]{TransactionHash}\n[BlockNumber]{BlockNumber}\n[GasUsed]{GasUsed}\n[Logs]{Logs}")
                DeployedContract = Contract(self, ContractAddress, ABI)
                return {"Status": Status, "TransactionHash": TransactionHash, "BlockNumber": BlockNumber, "GasUsed": GasUsed, "Logs": Logs, "ContractAddress": ContractAddress, "Contract": DeployedContract}
            else:
                logger.error(f"\n[ConfirmDeploy][Fail]\n[TransactionHash]{TransactionHash}")
                return {"Status": Status, "TransactionHash": TransactionHash}
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[DeployContract][Traditional]Failed to deploy contract.\n[ExceptionInformation]{ExceptionInformation}")
            return None

    def DeployContractByEIP1559(self, ABI: dict, Bytecode: str, Value: int = 0, *Arguments) -> dict:
        try:
            DeployingContract = self.Net.eth.contract(abi=ABI, bytecode=Bytecode)
            TransactionData = DeployingContract.constructor(*Arguments).buildTransaction({"value": Value})
            Txn = {
                "chainId": self.Net.eth.chainId,
                "from": self.Address,
                "nonce": self.Net.eth.get_transaction_count(self.Address),
                "value": TransactionData["value"],
                "maxFeePerGas": self.Net.eth.max_priority_fee * 2,
                "maxPriorityFeePerGas": round(self.Net.eth.max_priority_fee * 1.2),
                "gas": TransactionData["gas"],
                "data": TransactionData["data"]
            }
            SignedTxn = self.Net.eth.account.sign_transaction(Txn, self.PrivateKey)
            TransactionHash = self.Net.eth.send_raw_transaction(SignedTxn.rawTransaction).hex()
            logger.info(f"\n[DeployContract][EIP-1559]\n[TransactionHash]{TransactionHash}\n[Txn]{dumps(Txn, indent=2)}")
            TransactionReceipt = self.Net.eth.wait_for_transaction_receipt(TransactionHash, timeout=120)
            Status = TransactionReceipt.status
            if Status:
                ContractAddress = TransactionReceipt.contractAddress
                BlockNumber = TransactionReceipt.blockNumber
                GasUsed = TransactionReceipt.gasUsed
                Logs = TransactionReceipt.logs
                logger.success(f"\n[ConfirmDeploy][Success]\n[ContractAddress]{ContractAddress}\n[TransactionHash]{TransactionHash}\n[BlockNumber]{BlockNumber}\n[GasUsed]{GasUsed}\n[Logs]{Logs}")
                DeployedContract = Contract(self, ContractAddress, ABI)
                return {"Status": Status, "TransactionHash": TransactionHash, "BlockNumber": BlockNumber, "GasUsed": GasUsed, "Logs": Logs, "ContractAddress": ContractAddress, "Contract": DeployedContract}
            else:
                logger.error(f"\n[ConfirmDeploy][Fail]\n[TransactionHash]{TransactionHash}")
                return {"Status": Status, "TransactionHash": TransactionHash}
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[DeployContract][EIP-1559]Failed to deploy contract.\n[ExceptionInformation]{ExceptionInformation}")
            return None

    def SignMessage(self, Message: str) -> dict:
        from eth_account.messages import encode_defunct
        try:
            Temp = encode_defunct(text=Message)
            SignedMessage = self.Net.eth.account.sign_message(Temp, private_key=self.PrivateKey)
            SignedMessageHash = SignedMessage.messageHash.hex()
            SignedMessageSignature = SignedMessage.signature.hex()
            SignedMessageR = hex(SignedMessage.r)
            SignedMessageS = hex(SignedMessage.s)
            SignedMessageV = SignedMessage.v
            logger.success(
                f"\n[SignMessage][{self.Address}]\n[Message]{Message}\n[SignedMessageHash]{SignedMessageHash}\n[SignedMessageSignature]{SignedMessageSignature}\n[SignedMessageR]{SignedMessageR}\n[SignedMessageS]{SignedMessageS}\n[SignedMessageV]{SignedMessageV}")
            return {"Address": self.Address, "Message": Message, "SignedMessageHash": SignedMessageHash, "SignedMessageSignature": SignedMessageSignature, "SignedMessageR": SignedMessageR, "SignedMessageS": SignedMessageS, "SignedMessageV": SignedMessageV}
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[SignMessage]Failed to sign messge [{Message}] by account [{self.Address}].\n[ExceptionInformation]{ExceptionInformation}")
            return None


class Contract():
    def __init__(self, Account: Account, Address: str, ABI: dict):
        try:
            self.Account = Account
            self.Net = Account.Net
            self.Address = Web3.toChecksumAddress(Address)
            self.Instance = self.Net.eth.contract(address=self.Address, abi=ABI)
            logger.success(f"\n[InstantiateContract]Successfully instantiated contract [{self.Address}].")
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[InstantiateContract]Failed to instantiated contract [{self.Address}].\n[ExceptionInformation]{ExceptionInformation}")
            raise Exception("Failed to instantiate contract.")

    def CallFunction(self, Value: int, FunctionName: str, *FunctionArguments) -> dict:
        TransactionData = self.Instance.functions[FunctionName](*FunctionArguments).buildTransaction({"value": Value})
        logger.info(f"\n[CallFunction][Traditional][{self.Address}]\n[Function]{FunctionName}{FunctionArguments}\n[Value]{TransactionData['value']} [Gas]{TransactionData['gas']}")
        TransactionResult = self.Account.SendTransaction(self.Address, TransactionData["data"], TransactionData["value"], TransactionData["gas"])
        return TransactionResult

    def CallFunctionByEIP1559(self, Value: int, FunctionName: str, *FunctionArguments) -> dict:
        TransactionData = self.Instance.functions[FunctionName](*FunctionArguments).buildTransaction({"value": Value})
        logger.info(f"\n[CallFunction][EIP-1559][{self.Address}]\n[Function]{FunctionName}{FunctionArguments}\n[Value]{TransactionData['value']} [Gas]{TransactionData['gas']}")
        TransactionResult = self.Account.SendTransactionByEIP1559(self.Address, TransactionData["data"], TransactionData["value"], TransactionData["gas"])
        return TransactionResult

    def ReadOnlyCallFunction(self, FunctionName: str, *FunctionArguments):
        try:
            Result = self.Instance.functions[FunctionName](*FunctionArguments).call()
            logger.success(f"\n[CallFunction][ReadOnly][{self.Address}]\n[Function]{FunctionName}{FunctionArguments}\n[Result]{Result}")
            return Result
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[CallFunction][ReadOnly]Failed to call readonly function [{FunctionName}{FunctionArguments}].\n[ExceptionInformation]{ExceptionInformation}")
            return None

    def EncodeABI(self, FunctionName: str, *FunctionArguments) -> str:
        try:
            CallData = self.Instance.encodeABI(fn_name=FunctionName, args=FunctionArguments)
            logger.success(f"\n[EncodeABI]Success.\n[ContractAddress]{self.Address}\n[Function]{FunctionName}{FunctionArguments}\n[CallData]{CallData}")
            return CallData
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[EncodeABI]Failed to encode abi for [{FunctionName}{FunctionArguments}].\n[ExceptionInformation]{ExceptionInformation}")
            return None


class BlockchainUtils():
    @staticmethod
    def SwitchSolidityVersion(SolidityVersion: str):
        from solcx import install_solc, set_solc_version
        try:
            install_solc(SolidityVersion)
            set_solc_version(SolidityVersion)
            logger.success(f"\n[SwitchSolidityVersion]Current Version:{SolidityVersion}")
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[SwitchSolidityVersion]Failed to switch to version [{SolidityVersion}].\n[ExceptionInformation]{ExceptionInformation}")
            raise Exception("Failed to switch solidity version.")

    @staticmethod
    def Compile(FileCourse: str, ContractName: str, AllowPaths: str = None) -> tuple:
        from solcx import compile_source
        from json import dump
        try:
            with open(FileCourse, "r", encoding="utf-8") as sol:
                CompiledSol = compile_source(sol.read(), allow_paths=AllowPaths, optimize=True)
            ContractData = CompiledSol[f'<stdin>:{ContractName}']
            ABI = ContractData['abi']
            Bytecode = ContractData['bin']
            with open(f'{ContractName}_ABI.json', 'w') as f:
                dump(ABI, f, indent=4)
            logger.success(f"\n[CompileContract]Success.\n[FileCourse]{FileCourse}\n[ContractName]{ContractName}\n[ABI]{ABI}\n[Bytecode]{Bytecode}")
            return (ABI, Bytecode)
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[CompileContract]Failed to compile the contract [{FileCourse}][{ContractName}].\n[ExceptionInformation]{ExceptionInformation}")
            raise Exception("Failed to compile the contract.")

    @staticmethod
    def CreateNewAccount() -> tuple:
        from Crypto.Util.number import bytes_to_long
        Net = Web3()
        Keys = Net.eth.account.create()
        Address = Net.toChecksumAddress(Keys.address)
        PrivateKey = hex(bytes_to_long(Keys.privateKey))
        logger.success(f"\n[CreateNewAccount]\n[Address]{Address}\n[PrivateKey]{PrivateKey}")
        return (Address, PrivateKey)

    @staticmethod
    def RecoverMessage(Message: str, Signature: str) -> str:
        from eth_account.messages import encode_defunct
        try:
            Net = Web3()
            Temp = encode_defunct(text=Message)
            Signer = Net.eth.account.recover_message(Temp, signature=Signature)
            logger.success(f"\n[RecoverMessage]\n[Message]{Message}\n[Signature]{Signature}\n[Signer]{Signer}")
            return Signer
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[RecoverMessage]Failed to recover message [{Message}] with [{Signature}].\n[ExceptionInformation]{ExceptionInformation}")
            return None

    @staticmethod
    def RecoverMessageByHash(MessageHash: str, Signature: str) -> str:
        try:
            Net = Web3()
            Signer = Net.eth.account.recoverHash(MessageHash, signature=Signature)
            logger.success(f"\n[RecoverMessageByHash]\n[MessageHash]{MessageHash}\n[Signature]{Signature}\n[Signer]{Signer}")
            return Signer
        except Exception:
            ExceptionInformation = exc_info()
            logger.error(f"\n[RecoverMessageByHash]Failed to recover message hash [{MessageHash}] with [{Signature}].\n[ExceptionInformation]{ExceptionInformation}")
            return None



class MiscUtils():
    @staticmethod
    def Binary_String(Binary: str) -> str:
        assert(len(Binary) % 8 == 0)
        String = "".join([chr(int(Binary[i:i + 8], 2)) for i in range(0, len(Binary), 8)])
        return String

    @staticmethod
    def Binary_Dec(Binary: str) -> str:
        Dec = str(int(Binary, 2))
        return Dec

    @staticmethod
    def Binary_Hex(Binary: str) -> str:
        Hex = hex(int(Binary, 2))
        return Hex

    @staticmethod
    def Dec_String(Dec: int) -> str:
        from Crypto.Util.number import long_to_bytes
        String = long_to_bytes(Dec).decode()
        return String

    @staticmethod
    def Dec_Binary(Dec: int) -> bytes:
        Binary = bin(Dec)
        return Binary

    @staticmethod
    def Dec_Hex(Dec: int) -> str:
        Hex = hex(Dec)
        return Hex

    @staticmethod
    def Hex_String(Hex: str) -> str:
        assert(len(Hex) % 2 == 0)
        String = "".join([chr(int(Hex[i:i + 2], 16)) for i in range(0, len(Hex), 2)])
        return String

    @staticmethod
    def Hex_Binary(Hex: str) -> bytes:
        Binary = bin(int(Hex, 16))
        return Binary

    @staticmethod
    def Hex_Dec(Hex: str) -> str:
        Dec = str(int(Hex, 16))
        return Dec

    @staticmethod
    def SHA1(Text: str) -> str:
        from hashlib import sha1
        Hash = sha1(Text.encode()).hexdigest()
        return Hash

    @staticmethod
    def SHA256(Text: str) -> str:
        from hashlib import sha256
        Hash = sha256(Text.encode()).hexdigest()
        return Hash

    @staticmethod
    def SHA512(Text: str) -> str:
        from hashlib import sha512
        Hash = sha512(Text.encode()).hexdigest()
        return Hash

    @staticmethod
    def MD5(Text: str) -> str:
        from hashlib import md5
        Hash = md5(Text.encode()).hexdigest()
        return Hash


class MiscAdvancedUtils():
    @staticmethod
    def CRC_Burst(Course: str) -> None:
        from struct import pack
        from binascii import crc32
        with open(Course, "rb") as f:
            Binary = f.read()
        if(int(Binary[0:8].hex(), 16) != int("0x89504e470d0a1a0a", 16)):
            print("[CRC爆破]检测到[文件头错误]，请修改第[1-8]个字节为:89 50 4E 47 0D 0A 1A 0A")
        if(int(Binary[8:12].hex(), 16) != int("0x0000000d", 16)):
            print("[CRC爆破]检测到[数据块长度错误]，请修改第[9-12]个字节为:00 00 00 0D")
        if(int(Binary[12:16].hex(), 16) != int("0x49484452", 16)):
            print("[CRC爆破]检测到[数据块标识错误]，请修改第[13-16]个字节为:49 48 44 52")
        Code = crc32(Binary[12:29]) & 0xffffffff
        CRC = int(Binary[29:33].hex(), 16)
        if(Code == CRC):
            print("[CRC爆破]CRC校验码匹配")
        else:
            print("[CRC爆破]CRC校验码不匹配 即将进行[宽度]爆破")
            for i in range(1, 4097):
                temp = pack(">i", i)
                Code = crc32(bytes("IHDR", "ascii") + temp + Binary[20:29]) & 0xffffffff
                if(Code == CRC):
                    print("图片正确宽度为:{} Hex:{}".format(i, hex(i)))
                    return
            print("[CRC爆破][宽度]爆破失败 即将进行[高度]爆破")
            for i in range(1, 4097):
                temp = pack(">i", i)
                Code = crc32(bytes("IHDR", "ascii") + Binary[16:20] + temp + Binary[24:29]) & 0xffffffff
                if(Code == CRC):
                    print("图片正确高度为:{} Hex:{}".format(i, hex(i)))
                    return
            print("[CRC爆破][宽高单项爆破]失败 即将进行[宽高联合爆破]")
            for i in range(4097):
                width = pack(">i", i)
                for j in range(4097):
                    height = pack(">i", j)
                    Code = crc32(bytes("IHDR", "ascii") + width + height + Binary[24:29]) & 0xffffffff
                    if(Code == CRC):
                        print("[CRC爆破]宽度为:{} 高度为:{} Hex:{} {}".format(i, j, hex(i), hex(j)))
                        return
            print("[CRC爆破]爆破失败 请判断异常情况")

    @staticmethod
    def BinaryToQRCode(BinaryList: list) -> None:
        from PIL import Image
        X = Y = len(BinaryList)
        for i in range(X):
            assert(X == len(BinaryList[i]))
        image = Image.new('RGB', (X, Y))
        white = (255, 255, 255)
        black = (0, 0, 0)
        for i in range(X):
            line = BinaryList[i]
            for j in range(Y):
                if line[j] == '1':
                    image.putpixel((i, j), black)
                elif line[j] == '0':
                    image.putpixel((i, j), white)
        image.save("BinaryToQRCodeResult.png")
        image.show()

    @staticmethod
    def RGBToImage(RGBList: list, X: int, Y: int, Mode: str = "Column") -> None:
        from PIL import Image
        Board = Image.new("RGB", (X, Y))
        if Mode == "Column":
            for i in range(X):
                for j in range(Y):
                    Index = i * Y + j
                    Board.putpixel((i, j), tuple(RGBList[Index]))
        elif Mode == "Row":
            for i in range(Y):
                for j in range(X):
                    Index = i * X + j
                    Board.putpixel((j, i), tuple(RGBList[Index]))
        Board.save("RGBToImageResult.png")
        Board.show()

    @staticmethod
    def ImageToRGB(ImageCourse: str, Mode="Column") -> list:
        from PIL import Image
        image = Image.open(ImageCourse)
        X = image.size[0]
        Y = image.size[1]
        imageRGB = image.convert("RGB")
        Result = []
        if Mode == "Column":
            for i in range(X):
                for j in range(Y):
                    Result.append(imageRGB.getpixel((i, j)))
        elif Mode == "Row":
            for i in range(Y):
                for j in range(X):
                    Result.append(imageRGB.getpixel((j, i)))
        return Result

    @staticmethod
    def Request(Url: str, Method: str = "GET", Headers: dict = None, Params: dict = None, Data: dict = None) -> dict:
        from requests import request
        Response = request(Method, Url, headers=Headers, params=Params, data=Data)
        StatusCode = Response.status_code
        ResponseHeaders = Response.headers
        Cookies = Response.cookies
        Response.encoding = Response.apparent_encoding
        ResponseText = Response.text
        CompactResponseText = Response.text.replace("\n", "").replace(" ", "")
        Result = {"StatusCode": StatusCode, "ResponseHeaders": ResponseHeaders, "Cookies": Cookies, "ResponseText": ResponseText, "CompactResponseText": CompactResponseText}
        return Result



import gmpy2
import base64
from Crypto.Cipher import AES as aes
from Crypto.Util.number import long_to_bytes, bytes_to_long


class ModernCryptoUtils():
    @staticmethod
    def Base64_Encrypt(Text: str) -> str:
        try:
            EncryptedText = base64.b64encode(Text.encode()).decode()
            return EncryptedText
        except:
            return "[Base64]加密错误"

    @staticmethod
    def Base64_Decrypt(Text: str) -> str:
        try:
            DecryptedText = base64.b64decode(Text.encode()).decode()
            return DecryptedText
        except:
            return "[Base64]解密错误"

    @staticmethod
    def Base64_Stego_Decrypt(Base64List: list) -> str:
        Base64Code = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
        Binary = ""
        for line in Base64List:
            if "==" in line:
                temp = bin(Base64Code.find(line[-3]) & 15)[2:]  # 通过按位与&15运算取出二进制数后4位 [2:]的作用是将0b过滤掉
                Binary += "0" * (4 - len(temp)) + temp  # 高位补0
            elif "=" in line:
                temp = bin(Base64Code.find(line[-2]) & 3)[2:]  # 通过按位与&3运算取出二进制数后2位
                Binary += "0" * (2 - len(temp)) + temp  # 高位补0
        Text = ""
        if(len(Binary) % 8 != 0):  # 最终得到的隐写数据二进制位数不一定都是8的倍数，为了避免数组越界，加上一个判断
            print("[Base64隐写]将进行不完整解析")
            for i in range(0, len(Binary), 8):
                if(i + 8 > len(Binary)):
                    Text += " 剩余位:" + Binary[i:]
                    return Text
                else:
                    Text += chr(int(Binary[i:i + 8], 2))
        else:
            Text = "".join([chr(int(Binary[i:i + 8], 2)) for i in range(0, len(Binary), 8)])
            return Text

    @staticmethod
    def Base32_Encrypt(Text: str) -> str:
        try:
            EncryptedText = base64.b32encode(Text.upper().encode()).decode()
            return EncryptedText
        except:
            return "[Base32]加密错误"

    @staticmethod
    def Base32_Decrypt(Text: str) -> str:
        try:
            DecryptedText = base64.b32decode(Text.upper().encode()).decode()
            return DecryptedText
        except:
            return "[Base32]解密错误"

    @staticmethod
    def Base16_Encrypt(Text: str) -> str:
        try:
            EncryptedText = base64.b16encode(Text.upper().encode()).decode()
            return EncryptedText
        except:
            return "[Base16]加密错误"

    @staticmethod
    def Base16_Decrypt(Text: str) -> str:
        try:
            DecryptedText = base64.b16decode(Text.upper().encode()).decode()
            return DecryptedText
        except:
            return "[Base16]解密错误"

    @staticmethod
    def AES_Fill(Text: str) -> bytes:
        Fill = Text.encode()
        while len(Fill) % 16 != 0:
            Fill += b'\x00'
        return Fill

    @staticmethod
    def AES_Encrypt(Text: str, Key: str, Mode=aes.MODE_ECB) -> str:
        AES = aes.new(ModernCryptoUtils.AES_Fill(Key), Mode)
        EncryptedText = base64.b64encode(AES.encrypt(ModernCryptoUtils.AES_Fill(Text))).decode()
        return EncryptedText

    @staticmethod
    def AES_Decrypt(Text: str, Key: str, Mode=aes.MODE_ECB) -> str:
        AES = aes.new(ModernCryptoUtils.AES_Fill(Key), Mode)
        DecryptedText = AES.decrypt(base64.b64decode(ModernCryptoUtils.AES_Fill(Text))).decode()
        return DecryptedText

    @staticmethod
    def RSA_Encrypt(Text: str, p: int, q: int, e: int) -> str:
        m = bytes_to_long(Text.encode())
        c = gmpy2.powmod(m, e, p * q)
        EncryptedText = base64.b64encode(long_to_bytes(c)).decode()
        return EncryptedText

    @staticmethod
    def RSA_Base64_Decrypt(Base64: str, p: int, q: int, e: int) -> str:
        c = bytes_to_long(base64.b64decode(Base64.encode()))
        d = gmpy2.invert(e, (p - 1) * (q - 1))
        m = gmpy2.powmod(c, d, p * q)
        DecryptedText = long_to_bytes(m).decode()
        return DecryptedText

    @staticmethod
    def RSA_Long_Decrypt(Long: int, p: int, q: int, e: int) -> str:
        d = gmpy2.invert(e, (p - 1) * (q - 1))
        m = gmpy2.powmod(Long, d, p * q)
        DecryptedText = long_to_bytes(m).decode()
        return DecryptedText

    @staticmethod
    def RSA_Wiener_Attack(c: int, e: int, n: int) -> str:
        def continuedFra(x, y):
            """计算连分数
            :param x: 分子
            :param y: 分母
            :return: 连分数列表
            """
            cf = []
            while y:
                cf.append(x // y)
                x, y = y, x % y
            return cf

        def gradualFra(cf):
            """计算传入列表最后的渐进分数
            :param cf: 连分数列表
            :return: 该列表最后的渐近分数
            """
            numerator = 0
            denominator = 1
            for x in cf[::-1]:
                # 这里的渐进分数分子分母要分开
                numerator, denominator = denominator, x * denominator + numerator
            return numerator, denominator

        def solve_pq(a, b, c):
            """使用韦达定理解出pq，x^2−(p+q)∗x+pq=0
            :param a:x^2的系数
            :param b:x的系数
            :param c:pq
            :return:p，q
            """
            par = gmpy2.isqrt(b * b - 4 * a * c)
            return (-b + par) // (2 * a), (-b - par) // (2 * a)

        def getGradualFra(cf):
            """计算列表所有的渐近分数
            :param cf: 连分数列表
            :return: 该列表所有的渐近分数
            """
            gf = []
            for i in range(1, len(cf) + 1):
                gf.append(gradualFra(cf[:i]))
            return gf

        cf = continuedFra(e, n)
        gf = getGradualFra(cf)
        for d, k in gf:
            if k == 0:
                continue
            if (e * d - 1) % k != 0:
                continue
            phi = (e * d - 1) // k
            p, q = solve_pq(1, n - phi + 1, n)
            if p * q == n:
                m = long_to_bytes(gmpy2.powmod(c, d, n)).decode()
                return m

    @staticmethod
    def RSA_MultiPrime_Attack(c: int, e: int, n: int, primes: list, powers: list = None) -> str:
        from operator import mul
        from functools import reduce
        if powers == None:
            powers = [1 for _ in range(len(primes))]
        assert(len(primes) == len(powers))
        temp = [pow(primes[i], powers[i] - 1) * (primes[i] - 1) for i in range(len(primes))]
        phi = reduce(mul, temp)
        d = gmpy2.invert(e, phi)
        m = long_to_bytes(gmpy2.powmod(c, d, n)).decode()
        return m

    @staticmethod
    def RSA_LowEncryptionIndex_Attack(c: int, e: int, n: int) -> str:
        k = 0
        while True:
            m = gmpy2.iroot(c + k * n, e)[0]
            if type(m) == type(gmpy2.mpz(0)):
                m = long_to_bytes(m).decode()
                return m
            else:
                k += 1
                continue

    @staticmethod
    def RSA_CommonMod_Attack(c1: int, c2: int, e1: int, e2: int, n: int) -> str:
        _, x, y = gmpy2.gcdext(e1, e2)
        m = long_to_bytes((gmpy2.powmod(c1, x, n) * gmpy2.powmod(c2, y, n)) % n).decode()
        return m

    @staticmethod
    def RSA_Broadcast_Attack(cs: list, e: int, ns: list) -> str:
        from functools import reduce
        assert(len(cs) == len(ns))
        s = 0
        prod = reduce(lambda a, b: a * b, ns)
        for ni, ai in zip(ns, cs):
            p = prod / ni
            s += ai * gmpy2.invert(p, ni) * p
        x = s % prod
        m = long_to_bytes(gmpy2.iroot(x, e)).decode()
        return m

    @staticmethod
    def RC4_Encrypt(Text: str, Key: str) -> str:
        from Crypto.Cipher import ARC4
        RC4 = ARC4.new(Key.encode())
        EncryptedText = base64.b64encode(RC4.encrypt(Text.encode())).decode()
        return EncryptedText

    @staticmethod
    def RC4_Decrypt(Text: str, Key: str) -> str:
        from Crypto.Cipher import ARC4
        RC4 = ARC4.new(Key.encode())
        DecryptedText = RC4.decrypt(base64.b64decode(Text.encode())).decode()
        return DecryptedText


class ClassicalCryptoUtils():
    @staticmethod
    def Caesar_Encrypt(Text: str, Move: int = 3) -> str:
        EncryptedText = ""
        for i in Text:
            if i.isupper():
                EncryptedText += chr((ord(i) - ord('A') + int(Move)) % 26 + ord('A'))
            elif i.islower():
                EncryptedText += chr((ord(i) - ord('a') + int(Move)) % 26 + ord('a'))
            else:
                EncryptedText += i
        return EncryptedText

    @staticmethod
    def Caesar_Decrypt(Text: str, Move: int = 3) -> str:
        DecryptedText = ""
        for i in Text:
            if i.isupper():
                DecryptedText += chr((ord(i) - ord('A') - int(Move)) % 26 + ord('A'))
            elif i.islower():
                DecryptedText += chr((ord(i) - ord('a') - int(Move)) % 26 + ord('a'))
            else:
                DecryptedText += i
        return DecryptedText

    @staticmethod
    def Caesar_Attack(Text: str) -> list:
        Result = [ClassicalCryptoUtils.Caesar_Decrypt(Text, i) for i in range(0, 26)]
        return Result

    @staticmethod
    def Morse_Encrypt(Morse: str) -> str:
        MorseCode = {
            "A": ".-", "B": "-...", "C": "-.-.", "D": "-..", "E": ".", "F": "..-.", "G": "--.",
            "H": "....", "I": "..", "J": ".---", "K": "-.-", "L": ".-..", "M": "--",
            "N": "-.", "O": "---", "P": ".--.", "Q": "--.-", "R": ".-.", "S": "...",
            "T": "-", "U": "..-", "V": "...-", "W": ".--", "X": "-..-", "Y": "-.--",
            "Z": "--..", "a": ".-", "b": "-...", "c": "-.-.", "d": "-..", "e": ".",
            "f": "..-.", "g": "--.", "h": "....", "i": "..", "j": ".---", "k": "-.-",
            "l": ".-..", "m": "--", "n": "-.", "o": "---", "p": ".--.", "q": "--.-",
            "r": ".-.", "s": "...", "t": "-", "u": "..-", "v": "...-", "w": ".--",
            "x": "-..-", "y": "-.--", "z": "--..", "0": "-----", "1": ".----", "2": "..---",
            "3": "...--", "4": "....-", "5": ".....", "6": "-....", "7": "--...", "8": "---..",
            "9": "----.", ".": ".-.-.-", ":": "---...", ",": "--..--", ";": "-.-.-.", "?": "..--..",
            "=": "-...-", "'": ".----.", "/": "-..-.", "!": "-.-.--", "-": "-....-", "_": "..--.-",
            '"': ".-..-.", "(": "-.--.", ")": "-.--.-", "$": "...-..-", "&": ".-...", "@": ".--.-.", "+": ".-.-."
        }
        EncryptedText = "/".join([MorseCode.get(i, "[" + i + "]") for i in Morse])
        return EncryptedText

    @staticmethod
    def Morse_Decrypt(Morse: str) -> str:
        MorseCode = {
            ".-": "a", "-...": "b", "-.-.": "c", "-..": "d", ".": "e", "..-.": "f",
            "--.": "g", "....": "h", "..": "i", ".---": "j", "-.-": "k", ".-..": "l",
            "--": "m", "-.": "n", "---": "o", ".--.": "p", "--.-": "q", ".-.": "r",
            "...": "s", "-": "t", "..-": "u", "...-": "v", ".--": "w", "-..-": "x",
            "-.--": "y", "--..": "z", "-----": "0", ".----": "1", "..---": "2", "...--": "3",
            "....-": "4", ".....": "5", "-....": "6", "--...": "7", "---..": "8", "----.": "9",
            ".-.-.-": ".", "---...": ":", "--..--": ",", "-.-.-.": ";", "..--..": "?", "-...-": "=",
            ".----.": "'", "-..-.": "/", "-.-.--": "!", "-....-": "-", "..--.-": "_", ".-..-.": '"',
            "-.--.": "(", "-.--.-": ")", "...-..-": "$", ".-...": "&", ".--.-.": "@", ".-.-.": "+"
        }
        DecryptedText = "".join([MorseCode.get(i, "[" + i + "]") for i in Morse.split("/")])
        return DecryptedText

    @staticmethod
    def Bacon_Encrypt(Bacon: str) -> str:
        BaconCode = {
            "a": "aaaaa", "b": "aaaab", "c": "aaaba", "d": "aaabb", "e": "aabaa", "f": "aabab", "g": "aabba",
            "h": "aabbb", "i": "abaaa", "j": "abaab", "k": "ababa", "l": "ababb", "m": "abbaa", "n": "abbab",
            "o": "abbba", "p": "abbbb", "q": "baaaa", "r": "baaab", "s": "baaba", "t": "baabb",
            "u": "babaa", "v": "babab", "w": "babba", "x": "babbb", "y": "bbaaa", "z": "bbaab",
        }
        EncryptedText = "".join([BaconCode.get(i, "[" + i + "]") for i in Bacon.lower()]).upper()
        return EncryptedText

    @staticmethod
    def Bacon_Decrypt(Bacon: str) -> str:
        BaconCode = {
            "aaaaa": "a", "aaaab": "b", "aaaba": "c", "aaabb": "d", "aabaa": "e", "aabab": "f", "aabba": "g",
            "aabbb": "h", "abaaa": "i", "abaab": "j", "ababa": "k", "ababb": "l", "abbaa": "m", "abbab": "n",
            "abbba": "o", "abbbb": "p", "baaaa": "q", "baaab": "r", "baaba": "s", "baabb": "t",
            "babaa": "u", "babab": "v", "babba": "w", "babbb": "x", "bbaaa": "y", "bbaab": "z",
        }
        assert(len(Bacon) % 5 == 0)
        List = [Bacon.lower()[i:i + 5] for i in range(0, len(Bacon), 5)]
        DecryptText = "".join([BaconCode.get(i, "[" + i + "]") for i in List]).upper()
        return DecryptText

    @staticmethod
    def Vigenere_Encrypt(Text: str, Key: str) -> str:
        Key = Key[:len(Text)]
        Temp = Key
        index = 0
        while(len(Key) < len(Text)):
            if(index != len(Temp)):
                Key += Temp[index]
                index += 1
            else:
                index = 0
                Key += Temp[index]
                index += 1
        EncryptedText = ""
        for i in range(len(Text)):
            if Text[i].isupper():
                EncryptedText += chr((ord(Text[i]) + ord(Key[i].upper()) - 2 * ord('A')) % 26 + ord('A'))
            elif Text[i].islower():
                EncryptedText += chr((ord(Text[i]) + ord(Key[i].lower()) - 2 * ord('a')) % 26 + ord('a'))
            else:
                EncryptedText += Text[i]
        return EncryptedText

    @staticmethod
    def Vigenere_Decrypt(Text: str, Key: str) -> str:
        Key = Key[:len(Text)]
        Temp = Key
        index = 0
        while(len(Key) < len(Text)):
            if(index != len(Temp)):
                Key = Key + Temp[index]
                index = index + 1
            else:
                index = 0
                Key = Key + Temp[index]
                index = index + 1
        DecryptedText = ""
        for i in range(len(Text)):
            if Text[i].isupper():
                DecryptedText += chr((ord(Text[i]) - ord(Key[i].upper())) % 26 + ord('A'))
            elif Text[i].islower():
                DecryptedText += chr((ord(Text[i]) - ord(Key[i].lower())) % 26 + ord('a'))
            else:
                DecryptedText += Text[i]
        return DecryptedText

    @staticmethod
    def Fence_Encrypt(Text: str, Fence: int) -> str:
        assert(len(Text) % Fence == 0)
        templist = []
        List = [Text[i:i + Fence] for i in range(0, len(Text), Fence)]
        for i in range(Fence):
            temp = ""
            for j in List:
                temp += j[i]
            templist.append(temp)
        EncryptedText = "".join(templist)
        return EncryptedText

    @staticmethod
    def Fence_Decrypt(Text: str, Fence: int) -> str:
        assert(len(Text) % Fence == 0)
        templist = ["" for _ in range(Fence)]
        List = [Text[i:i + Fence] for i in range(0, len(Text), Fence)]
        for i in List:
            for j in range(Fence):
                templist[j] = templist[j] + i[j]
        DecryptedText = "".join(templist)
        return DecryptedText

    @staticmethod
    def Fence_Attack(Text: str) -> list:
        Factors = [factor for factor in range(2, len(Text)) if len(Text) % factor == 0]
        Result = [(ClassicalCryptoUtils.Fence_Decrypt(Text, i), i) for i in Factors]
        return Result

    @staticmethod
    def WFence_Generate(Text: str, Fence: int) -> list:
        Matrix = [['.'] * len(Text) for _ in range(Fence)]
        Row = 0
        Up = False
        for Column in range(len(Text)):
            Matrix[Row][Column] = Text[Column]
            if Row == Fence - 1:
                Up = True
            if Row == 0:
                Up = False
            if Up:
                Row -= 1
            else:
                Row += 1
        return Matrix

    @staticmethod
    def WFence_Encrypt(Text: str, Fence: int) -> str:
        Matrix = ClassicalCryptoUtils.WFence_Generate(Text, Fence)
        EncryptedText = ""
        for Row in range(Fence):
            for Column in range(len(Text)):
                if Matrix[Row][Column] != '.':
                    EncryptedText += Matrix[Row][Column]
        return EncryptedText

    @staticmethod
    def WFence_Decrypt(Text: str, Fence: int) -> str:
        Matrix = ClassicalCryptoUtils.WFence_Generate(Text, Fence)
        index = 0
        for Row in range(Fence):
            for Column in range(len(Text)):
                if Matrix[Row][Column] != '.':
                    Matrix[Row][Column] = Text[index]
                    index += 1
        DecryptedText = ""
        for Column in range(len(Text)):
            for Row in range(Fence):
                if Matrix[Row][Column] != '.':
                    DecryptedText += Matrix[Row][Column]
        return DecryptedText

    @staticmethod
    def WFence_Attack(Text: str) -> list:
        Result = [(ClassicalCryptoUtils.WFence_Decrypt(Text, i), i) for i in range(2, len(Text))]
        return Result

    @staticmethod
    def Affine_Encrypt(Text: str, a: int, b: int) -> str:
        A = [1, 3, 5, 7, 9, 11, 15, 17, 19, 21, 23, 25]
        assert(a in A)
        List = []
        for i in Text:
            if i.isupper():
                List.append(chr((a * (ord(i) - ord('A')) + b) % 26 + ord('A')))
            elif i.islower():
                List.append(chr((a * (ord(i) - ord('a')) + b) % 26 + ord('a')))
            else:
                List.append(i)
        EncryptedText = "".join(List)
        return EncryptedText

    @staticmethod
    def Affine_Decrypt(Text: str, a: int, b: int) -> str:
        A = [1, 3, 5, 7, 9, 11, 15, 17, 19, 21, 23, 25]
        assert(a in A)
        List = []
        n = 1
        while (n * a) % 26 != 1:
            n = n + 1
        for i in Text:
            if i.isupper():
                List.append(chr(n * (ord(i) - ord('A') - b) % 26 + ord('A')))
            elif i.islower():
                List.append(chr(n * (ord(i) - ord('a') - b) % 26 + ord('a')))
            else:
                List.append(i)
        DecryptedText = "".join(List)
        return DecryptedText

    @staticmethod
    def Affine_Attack(Text: str) -> list:
        A = [1, 3, 5, 7, 9, 11, 15, 17, 19, 21, 23, 25]
        Result = []
        for a in A:
            for b in range(26):
                Result.append((ClassicalCryptoUtils.Affine_Decrypt(Text, a, b), a, b))
        return Result

    @staticmethod
    def Zodiac(Text: str, Foot: int) -> str:
        assert(len(Text) % Foot == 0)
        templist1 = [Text[i:i + Foot] for i in range(0, len(Text), Foot)]
        templist2 = []
        for index in range(Foot):
            temp = ""
            for i in range(0, len(templist1)):
                temp += templist1[i][index]
                index = (index + 2) % Foot
            templist2.append(temp)
        DecryptedText = "".join(templist2)
        return DecryptedText

    @staticmethod
    def Yunying(Text: str) -> str:
        List = Text.split("0")
        DecryptedText = ""
        for i in List:
            Sum = 64
            for j in i:
                Sum += int(j)
            DecryptedText += chr(Sum)
        return DecryptedText



from setuptools import setup

setup(
    name="poseidon-python",
    version="1.0.3",
    author="B1ue1nWh1te",
    author_email="b1ue1nwh1te@skiff.com",
    description="CTF解题快速利用工具，是攻克Blockchain方向的得力助手，也包含一些Crypto和Misc方向的小功能，可用于快速编写解题脚本而无需繁琐的步骤。",
    keywords=['POSEIDON', 'BLOCKCHAIN', 'ETHEREUM', 'CTF-TOOLS'],
    license="GPL-3.0",
    url="https://github.com/B1ue1nWh1te/Poseidon",
    python_requires='>=3.7',
    install_requires=["web3", "py-solc-x", "loguru", "pycryptodome", "gmpy2", "pwntools", "requests", "Pillow"],
    packages=["Poseidon"],
    zip_safe=False
)



import hashlib
from pwn import *
from pwnlib.util.iters import mbruteforce


class PoWUtils():
    @staticmethod
    def ProofOfWork_SHA256_Full(Url: str, Port: int, Length: int, HashBegin: str, SendAfter: str) -> remote:
        Connection = remote(Url, Port)
        Connection.recvuntil(HashBegin)
        Hash = Connection.recvuntil('\n', drop=True).decode().strip()
        Charset = string.printable
        Proof = mbruteforce(lambda x: hashlib.sha256((x).encode()).hexdigest() == Hash, Charset, Length, method='fixed')
        Connection.sendlineafter(SendAfter, Proof)
        return Connection

    @staticmethod
    def ProofOfWork_SHA256_Prefix(Url: str, Port: int, PrefixLength: int, HashBegin: str, SendAfter: str) -> remote:
        Connection = remote(Url, Port)
        Connection.recvuntil(HashBegin)
        Prefix = Connection.recvuntil('\n', drop=True).decode().strip()
        Charset = string.printable
        Proof = mbruteforce(lambda x: hashlib.sha256((x).encode()).hexdigest()[:PrefixLength] == Prefix, Charset, 8, method='upto')
        Connection.sendlineafter(SendAfter, Proof)
        return Connection

    @staticmethod
    def ProofOfWork_MD5_Full(Url: str, Port: int, Length: int, HashBegin: str, SendAfter: str) -> remote:
        Connection = remote(Url, Port)
        Connection.recvuntil(HashBegin)
        Hash = Connection.recvuntil('\n', drop=True).decode().strip()
        Charset = string.printable
        Proof = mbruteforce(lambda x: hashlib.md5((x).encode()).hexdigest() == Hash, Charset, Length, method='fixed')
        Connection.sendlineafter(SendAfter, Proof)
        return Connection

    @staticmethod
    def ProofOfWork_MD5_Prefix(Url: str, Port: int, PrefixLength: int, HashBegin: str, SendAfter: str) -> remote:
        Connection = remote(Url, Port)
        Connection.recvuntil(HashBegin)
        Prefix = Connection.recvuntil('\n', drop=True).decode().strip()
        Charset = string.printable
        Proof = mbruteforce(lambda x: hashlib.md5((x).encode()).hexdigest()[:PrefixLength] == Prefix, Charset, 8, method='upto')
        Connection.sendlineafter(SendAfter, Proof)
        return Connection
